#include "TimeRangeInterface.h"

#include "CallPermissions.h"

CallPermissions::CallPermissions()
{
	try {
		boost::property_tree::ptree pt;
		boost::property_tree::json_parser::read_json("/etc/asterisk/kerio_callPermissions.conf", pt);

		/* CallPermissions */
		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_callPermissions.callPermissions"))
		{
			CallPermissionsEntry cp(val.second);
			callPermissionList.insert(std::pair<int, CallPermissionsEntry > (cp.groupId, cp));
		}

		/* Extensions */
		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_callPermissions.extensions"))
		{
			CallPermissionsExtension ext(val.second);
			extensionList.insert(std::pair<std::string, CallPermissionsExtension > (ext.telNum, ext));
		}

		/* Interfaces */
		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_callPermissions.interfaces"))
		{
			CallPermissionsInterface iface(val.second);
			interfaceList.insert(std::pair<std::string, CallPermissionsInterface > (iface.name, iface));
		}

		/* Routes */
		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_callPermissions.routes"))
		{
			routeList.insert(val.second.data());
		}

		/* Emergency */
		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_callPermissions.emergency"))
		{
			emergencyList.insert(val.second.data());
		}
		bool emergencyDirectDialing = pt.get<bool >("kerio_callPermissions.emergencyDirectDialing");
		if (emergencyDirectDialing) {
			/* add empty prefix to the route list */
			routeList.insert("");
		}

		/* DisabledExtensions */
		BOOST_FOREACH(boost::property_tree::ptree::value_type &val, pt.get_child("kerio_callPermissions.disabledExtensions"))
		{
			CallPermissionsDisabledExtension de(val.second);
			disabledExtensionList.insert(std::pair<std::string, CallPermissionsDisabledExtension > (de.telNum, de));
		}
	}
	catch (std::exception & e) {
		ast_log(AST_LOG_ERROR, "Configuration error %s\n", e.what());
	}
}

CallPermissions::~CallPermissions() { }

/**
 * Test permissions of a call
 *
 * @param source [std::string] - extension to test permissions on
 * @param dialedNumber [std::string] - dialed number to test
 *
 * @return [int] - -1 on error, 0 on deny, 1 on permit
 */
int CallPermissions::testPermissions(const std::string source, const std::string dialedNumber)
{
	std::string groupName;
	int result = -1;

	/* test dialedNumber against emergency numbers */
	BOOST_FOREACH(const std::string &routePrefix, routeList)
	{

		BOOST_FOREACH(const std::string &emergencyNumber, emergencyList)
		{
			if (dialedNumber == routePrefix + emergencyNumber) {
				return 1; /* permit */
			}
		}
	}

	if (0 == source.compare(0, 10, "extension-")) {
		std::string extensionNumber = source.substr(10);

		/* test permissions */
		result = this->testPermissionsExtension(groupName, extensionNumber, dialedNumber);
		if (0 == result) {
			ast_log(AST_LOG_WARNING, "[security] Call Permissions: Call to \"%s\" from \"%s\" is forbidden by Call Permission Group '%s'.\n", dialedNumber.c_str(), extensionNumber.c_str(), groupName.c_str());
			return 0; /* deny */
		}

		/* test disabled user - source number */
		DisabledExtensionList::iterator itDisabled = disabledExtensionList.find(extensionNumber);
		if (disabledExtensionList.end() != itDisabled) {
			/* found in the list */
			if (itDisabled->second.disabled) {
				ast_log(AST_LOG_WARNING, "[security] Call Permissions: Call from extension \"%s\" is forbidden because User \"%s\" is disabled.\n", extensionNumber.c_str(), itDisabled->second.username.c_str());
				return 0; /* deny */
			}
		}
	}
	else if (0 == source.compare(0, 10, "interface-")) {
		std::string interface = source.substr(10);
		result = this->testPermissionsInterface(groupName, interface, dialedNumber);
		if (0 == result) {
			ast_log(AST_LOG_WARNING, "[security] Call Permissions: Call to \"%s\" originated at interface \"%s\" is forbidden by Call Permission Group '%s'.\n", dialedNumber.c_str(), interface.c_str(), groupName.c_str());
			return 0; /* deny */
		}
	}
	else if (source == "disabled") {
		/* disabled users only, it will be checked later */
		result = 1; /* permit */
	}
	else {
		ast_log(AST_LOG_WARNING, "Unknown source %s.\n", source.c_str());
		result = -1; /* error */
	}

	/* test disabled user - dialed number */
	DisabledExtensionList::iterator itDisabled = disabledExtensionList.find(dialedNumber);
	if (disabledExtensionList.end() != itDisabled) {
		/* found in the list */
		if (itDisabled->second.disabled) {
			ast_log(AST_LOG_WARNING, "[security] Call Permissions: Call to extension \"%s\" is forbidden because user \"%s\" is disabled.\n", dialedNumber.c_str(), itDisabled->second.username.c_str());
			return 0; /* deny */
		}
	}

	return result;
}

int CallPermissions::testPermissionsInterface(std::string &groupName, const std::string interfaceName, const std::string dialedNumber)
{
	/* test dialedNumber against call permissions */
	InterfaceList::iterator itInterface = interfaceList.find(interfaceName);
	if (interfaceList.end() == itInterface) {
		/* interface not found */
		ast_log(AST_LOG_WARNING, "Unknown interface %s.\n", interfaceName.c_str());
		return -1; /* error */
	}

	/* get subset from multimap */
	std::pair<std::multimap<int, CallPermissionsEntry>::iterator, std::multimap<int, CallPermissionsEntry>::iterator> groupPermissions
			= callPermissionList.equal_range(itInterface->second.groupId);

	BOOST_FOREACH(CallPermissionList::value_type it, groupPermissions)
	{
		CallPermissionsEntry perm = it.second;
		groupName = perm.groupName;

		/* test time range only when specified */
		if (0 < perm.timeRangeName.size()) {
			if (1 != timerange_test(perm.timeRangeName.c_str())) {
				continue; /* timerange miss */
			}
		}

		/* test for default action */
		if ("default" == perm.prefix) {
			return perm.policy;
		}

		/* test prefix */
		if (0 == dialedNumber.find(perm.prefix)) {
			return perm.policy;
		}
	}

	ast_log(AST_LOG_WARNING, "No policy matched for dialing %s originated from %s (group %d).\n", dialedNumber.c_str(), interfaceName.c_str(), itInterface->second.groupId);
	return -1; /* error */
}

int CallPermissions::testPermissionsExtension(std::string &groupName, const std::string extensionNumber, const std::string dialedNumber)
{
	/* test dialedNumber against call permissions */
	ExtensionList::iterator itExtension = extensionList.find(extensionNumber);
	if (extensionList.end() == itExtension) {
		/* extension not found */
		ast_log(AST_LOG_WARNING, "Unknown extension %s.\n", extensionNumber.c_str());
		return -1; /* error */
	}

	/* get subset from multimap */
	std::pair<std::multimap<int, CallPermissionsEntry>::iterator, std::multimap<int, CallPermissionsEntry>::iterator> groupPermissions
			= callPermissionList.equal_range(itExtension->second.groupId);

	BOOST_FOREACH(CallPermissionList::value_type it, groupPermissions)
	{
		CallPermissionsEntry perm = it.second;
		groupName = perm.groupName;

		/* test time range only when specified */
		if (0 < perm.timeRangeName.size()) {
			if (1 != timerange_test(perm.timeRangeName.c_str())) {
				continue; /* timerange miss */
			}
		}

		/* test for default action */
		if ("default" == perm.prefix) {
			return perm.policy;
		}

		/* test prefix */
		if (0 == dialedNumber.find(perm.prefix)) {
			return perm.policy;
		}
	}

	ast_log(AST_LOG_WARNING, "No policy matched for dialing %s on extension %s (group %d).\n", dialedNumber.c_str(), extensionNumber.c_str(), itExtension->second.groupId);
	return -1; /* error */
}
