#include "Interface.h"
#include "CallPermissions.h"
#include "CallPermissionsInterface.h"

static CallPermissions *callPermissions;
static Lock loadLock;

int callpermissions_load()
{
	LockUnique l(&loadLock.l);
	if (!l.ok) return 0;

	callPermissions = new CallPermissions();
	return 0;
}

void callpermissions_unload()
{
	LockUnique l(&loadLock.l);
	if (!l.ok) return;

	if (callPermissions) {
		delete callPermissions;
		callPermissions = NULL;
	}
}

void callpermissions_print_permissions(const int fd)
{
	LockShared l(&loadLock.l);
	if (!callPermissions) {
		return;
	}

	int lastGroupId = -1;

	ast_cli(fd, "Permissions list\n");
	ast_cli(fd, "-----------------------------------\n");

	BOOST_FOREACH(CallPermissions::CallPermissionList::value_type entry, callPermissions->callPermissionList)
	{
		int groupId = entry.first;
		CallPermissionsEntry perm = entry.second;

		if (lastGroupId != groupId) {
			ast_cli(fd, "Group %d - \"%s\"\n", groupId, perm.groupName.c_str());
			lastGroupId = groupId;
		}

		if (0 == perm.policy) {
			ast_cli(fd, "    [Deny]  ");
		}
		else {
			ast_cli(fd, "    [Allow] ");
		}

		ast_cli(fd, "\"%s\"", perm.prefix.c_str());

		if (0 < perm.timeRangeName.size()) {
			ast_cli(fd, " (\"%s\")", perm.timeRangeName.c_str());
		}

		ast_cli(fd, "\n");
	}
	ast_cli(fd, "-----------------------------------\n");
}

void callpermissions_print_extensions(int fd)
{
	LockShared l(&loadLock.l);
	if (!callPermissions) {
		return;
	}

	ast_cli(fd, "Extensions list\n");
	ast_cli(fd, "-----------------------------------\n");
	ast_cli(fd, "Extension number | Permission group\n");

	BOOST_FOREACH(CallPermissions::ExtensionList::value_type ext, callPermissions->extensionList)
	{
		ast_cli(fd, " %15s | %d\n", ext.second.telNum.c_str(), ext.second.groupId);
	}
	ast_cli(fd, "-----------------------------------\n");
}

void callpermissions_print_interfaces(int fd)
{
	LockShared l(&loadLock.l);
	if (!callPermissions) {
		return;
	}

	ast_cli(fd, "Interface list\n");
	ast_cli(fd, "---------------------------------------------\n");
	ast_cli(fd, "Interface                  | Permission group\n");

	BOOST_FOREACH(CallPermissions::InterfaceList::value_type iface, callPermissions->interfaceList)
	{
		ast_cli(fd, " %25s | %d\n", iface.second.name.c_str(), iface.second.groupId);
	}
	ast_cli(fd, "---------------------------------------------\n");
}

void callpermissions_print_emergency(int fd)
{
	LockShared l(&loadLock.l);
	if (!callPermissions) {
		return;
	}

	ast_cli(fd, "Emergency list\n");
	ast_cli(fd, "--------------\n");

	BOOST_FOREACH(CallPermissions::EmergencyList::value_type num, callPermissions->emergencyList)
	{
		ast_cli(fd, " %13s\n", num.c_str());
	}
	ast_cli(fd, "--------------\n");
}

void callpermissions_print_routes(int fd)
{
	LockShared l(&loadLock.l);
	if (!callPermissions) {
		return;
	}

	ast_cli(fd, "Route list\n");
	ast_cli(fd, "--------------\n");

	BOOST_FOREACH(CallPermissions::RouteList::value_type num, callPermissions->routeList)
	{
		ast_cli(fd, " %13s\n", num.c_str());
	}
	ast_cli(fd, "--------------\n");
}

void callpermissions_print_disabled(int fd)
{
	LockShared l(&loadLock.l);
	if (!callPermissions) {
		return;
	}

	ast_cli(fd, "Disabled user list\n");
	ast_cli(fd, "------------------\n");

	BOOST_FOREACH(CallPermissions::DisabledExtensionList::value_type ext, callPermissions->disabledExtensionList)
	{
		ast_cli(fd, " %13s %13s - %s\n", ext.second.telNum.c_str(), ext.second.username.c_str(), ext.second.disabled ? "disabled" : "enabled");
	}
	ast_cli(fd, "------------------\n");
}

void callpermissions_print(const int fd)
{
	LockShared l(&loadLock.l);
	if (!callPermissions) {
		return;
	}

	callpermissions_print_permissions(fd);
	ast_cli(fd, "\n");
	callpermissions_print_extensions(fd);
	ast_cli(fd, "\n");
	callpermissions_print_interfaces(fd);
	ast_cli(fd, "\n");
	callpermissions_print_emergency(fd);
	ast_cli(fd, "\n");
	callpermissions_print_routes(fd);
	ast_cli(fd, "\n");
	callpermissions_print_disabled(fd);
}

int callpermissions_test(const char *source, const char *dialedNumber)
{
	LockShared l(&loadLock.l);
	if (!callPermissions) {
		return 0;
	}

	/* call permissions */
	int status = callPermissions->testPermissions(source, dialedNumber);
	switch (status) {
		case 0: //miss
			ast_verb(3, "Call permissions for %s to dial %s... [DENY]\n", source, dialedNumber);
			break;
		case 1: //hit
			ast_verb(3, "Call permissions for %s to dial %s... [ALLOW]\n", source, dialedNumber);
			break;
		case -1: // invalid parameter
			ast_log(AST_LOG_WARNING, "Call permissions for %s to dial %s... [UNKNOWN]\n", source, dialedNumber);
			break;
	}

	return status;
}
