#include "Security.h"

Security::Security() : loaded(false), ruleSipAttackCount(0), ruleDuration(0), actionBlockDuration(0), actionEmailAlertEnabled(0)
{
	/* open sqlite */
	if (!this->sqlite.open(OPERATOR_RuntimeDb)) {
		ast_log(AST_LOG_ERROR, "Security Database connection error.\n");
		return;
	}

	try {
		boost::property_tree::ptree pt;
		boost::property_tree::json_parser::read_json("/etc/asterisk/kerio_security.conf", pt);

		this->ruleSipAttackCount = pt.get<int >("kerio_security.ruleSipAttackCount");
		this->ruleDuration = pt.get<int >("kerio_security.ruleDuration");
		this->actionBlockDuration = pt.get<int >("kerio_security.actionBlockDuration");
		this->actionEmailAlertEnabled = pt.get<bool >("kerio_security.actionEmailAlertEnabled");
	}
	catch (std::exception & e) {
		ast_log(AST_LOG_ERROR, "Configuration error %s\n", e.what());
		return;
	}

	this->loaded = true;
}

Security::~Security()
{
	/* close sqlite */
	this->sqlite.close();
	this->loaded = false;
}

void Security::print(const int fd)
{
	if (!this->loaded) {
		ast_cli(fd, "Security not loaded.\n");
		return;
	}
	LockUnique sqliteGuard(&this->sqliteLock.l);

	ast_cli(fd, "Security authentication failures:\n");
	ast_cli(fd, "------------------------------------------\n");
	ast_cli(fd, "%15s | %24s\n", "IP ADDRESS", "TIME");

	SqliteResult result;
	if (this->sqlite.query(result, "SELECT timestamp, ipAddress FROM SecurityLoginFailures ORDER BY timestamp ASC")) {
		while (this->sqlite.fetchRow(result)) {
			struct in_addr ipAddress;
			char strTime[256];
			struct tm *timeinfo;
			time_t rawtime;

			/* parse time */
			rawtime = sqlite3_column_int(result.statement, 0);
			timeinfo = localtime(&rawtime);
			strncpy(strTime, asctime(timeinfo), sizeof (strTime));
			strTime[strlen(strTime) - 1] = '\0'; /* remove trailing \n */

			/* parse ip address */
			ipAddress.s_addr = htonl(sqlite3_column_int(result.statement, 1));

			ast_cli(fd, "%15s | %24s\n", ast_inet_ntoa(ipAddress), strTime);
		}
	}
	ast_cli(fd, "------------------------------------------\n");
	ast_cli(fd, "\n");

	ast_cli(fd, "Security blocked ip addresses:\n");
	ast_cli(fd, "---------------------------------------------------------------------\n");
	ast_cli(fd, "%15s | %24s | %24s\n", "IP ADDRESS", "BLOCKED FROM", "BLOCKED TO");

	if (this->sqlite.query(result, "SELECT blockedFrom, blockedTo, ipAddress FROM SecurityBlockedIp ORDER BY ipAddress ASC")) {
		while (this->sqlite.fetchRow(result)) {
			struct in_addr ipAddress;
			char strBlockedFrom[256];
			char strBlockedTo[256];
			struct tm *timeinfo;
			time_t rawtime;

			/* parse blockedFrom */
			rawtime = sqlite3_column_int(result.statement, 0);
			timeinfo = localtime(&rawtime);
			strncpy(strBlockedFrom, asctime(timeinfo), sizeof (strBlockedFrom));
			strBlockedFrom[strlen(strBlockedFrom) - 1] = '\0'; /* remove trailing \n */

			/* parse blockedTo */
			rawtime = sqlite3_column_int(result.statement, 1);
			timeinfo = localtime(&rawtime);
			strncpy(strBlockedTo, asctime(timeinfo), sizeof (strBlockedTo));
			strBlockedTo[strlen(strBlockedTo) - 1] = '\0'; /* remove trailing \n */

			/* parse ip address */
			ipAddress.s_addr = htonl(sqlite3_column_int(result.statement, 2));

			ast_cli(fd, "%15s | %24s | %24s\n", ast_inet_ntoa(ipAddress), strBlockedFrom, strBlockedTo);
		}
	}
	ast_cli(fd, "---------------------------------------------------------------------\n");
	ast_cli(fd, "\n");
}

/**
 * Remove old entries from SecurityLoginFailures and SecurityBlockedIp
 */
void Security::updateLoginFailures()
{
	if (!this->loaded) {
		ast_log(AST_LOG_WARNING, "Security not loaded.\n");
		return;
	}

	time_t now = time(NULL);
	time_t t = now - ruleDuration;
	this->sqlite.exec("DELETE FROM SecurityLoginFailures WHERE timestamp < %d", (int) t);
	this->sqlite.exec("DELETE FROM SecurityBlockedIp WHERE blockedTo < %d", (int) now);
}

/**
 * Send an email notification
 *
 * @param sin
 * @param failureCount
 * @param tNow
 * @param tTo
 */
void Security::emailNotification(const struct sockaddr_in *sin, const int failureCount, const time_t tNow, const time_t tTo)
{
	if (!actionEmailAlertEnabled) {
		return;
	}

	ast_log(AST_LOG_NOTICE, "Sending email notification (blocked host %s)...\n", ast_inet_ntoa(sin->sin_addr));

	char tmp[1024];
	snprintf(tmp, sizeof (tmp), "%s newBlockedIp --IP_ADDRESS \"%s\" --FAILURE_COUNT %d &",
			mailCommand,
			ast_inet_ntoa(sin->sin_addr),
			failureCount
			);

	ast_safe_system(tmp);

	ast_log(AST_LOG_NOTICE, "Sending email notification (blocked host %s)... done.\n", ast_inet_ntoa(sin->sin_addr));
}

/**
 * Find ip address in SecurityBlockedIp
 *  1. delete old entries
 *  2. find ipAddress
 *      - when found extend the blocking period
 *
 * This function is called from chan_sip
 *
 * @param sin
 * @param user - format "ext@ip"
 * @return [int] 1 - allow, 0 - deny
 */
int Security::authCheck(const struct sockaddr_in *sin, const char *user)
{
	if (!this->loaded) {
		ast_log(AST_LOG_WARNING, "Security not loaded.\n");
		return 0; /* deny */
	}

	/* Allow traffic from 127.0.0.1 */
	if (sin->sin_addr.s_addr == 16777343 /* (uint32_t) inet_addr("127.0.0.1") */) {
		return 1; /* allow */
	}

	LockUnique sqliteGuard(&this->sqliteLock.l);

	time_t now = time(NULL);

	ast_debug(2, "KerioSecurity check %s (uri:%s).\n", ast_inet_ntoa(sin->sin_addr), user);

	/* find ipAddress */
	SqliteResult result;
	if (this->sqlite.query(result, "SELECT blockedFrom FROM SecurityBlockedIp WHERE ipAddress = %u AND %d < blockedTo", htonl(sin->sin_addr.s_addr), (int) now)) {
		if (this->sqlite.fetchRow(result)) {
			time_t blockedFrom = sqlite3_column_int(result.statement, 0);
			result.reset();

			/* do not update the table until at least half of the actionBlockDuration has passed */
			/* Bug 92340 - Asterisk is killed by watchdog during a DOS attack */
			if (now - blockedFrom <= (actionBlockDuration / 2)) {
				return 0; /* deny */
			}

			/* extend the blocking period */
			time_t newBlockingTime = now + actionBlockDuration;
			this->sqlite.exec("UPDATE SecurityBlockedIp SET blockedTo = %d WHERE ipAddress = %u",
					(int) newBlockingTime,
					htonl(sin->sin_addr.s_addr)
					);
			firewallBlockAddress(sin);

			return 0; /* deny */
		}
	}
	result.reset();

	return 1 /* allow */;
}

/**
 * Updates database when authentication failed
 *  1. add entry to the SecurityLoginFailures
 *  2. update SecurityBlockedIp when the conditions are met & email notify
 *
 * This function is called from chan_sip
 *
 * @param sin
 * @param user
 */
void Security::authFailure(const struct sockaddr_in *sin, const char *user)
{
	if (!this->loaded) {
		ast_log(AST_LOG_WARNING, "Security not loaded.\n");
		return;
	}

	/* Allow traffic from 127.0.0.1 */
	if (sin->sin_addr.s_addr == 16777343 /* (uint32_t) inet_addr("127.0.0.1") */) {
		return;
	}

	LockUnique sqliteGuard(&this->sqliteLock.l);

	int failureCount = 0;
	time_t tNow = time(NULL);

	/* update SecurityLoginFailures */
	updateLoginFailures();

	/* add new entry to SecurityLoginFailures */
	this->sqlite.exec("INSERT INTO SecurityLoginFailures (timestamp, ipAddress) VALUES (%d, %u)",
			(int) tNow,
			htonl(sin->sin_addr.s_addr)
			);

	/* update SecurityBlockedIp */
	SqliteResult result;
	if (this->sqlite.query(result, "SELECT count(*) FROM SecurityLoginFailures WHERE ipAddress=%u", htonl(sin->sin_addr.s_addr))) {
		if (this->sqlite.fetchRow(result)) {
			failureCount = sqlite3_column_int(result.statement, 0);
		}
		else {
			ast_log(AST_LOG_ERROR, "SecurityModule_sqlite: Database fetch error.\n");
		}
	}
	result.reset();

	ast_log(AST_LOG_WARNING, "[security] SIP: Authentication failed from host %s (uri:%s). Attempts: %d.\n", ast_inet_ntoa(sin->sin_addr), user, failureCount);

	/* add new entry to SecurityBlockedIp when ruleSipAttackCount reached */
	if (failureCount >= ruleSipAttackCount) {
		time_t tTo = tNow + actionBlockDuration;
		struct tm *timeinfo;
		timeinfo = localtime(&tTo);
		ast_log(AST_LOG_WARNING, "[security] SIP: Too many failed authentication attempts (%d) from host %s. Blocking host until: %s\n", failureCount, ast_inet_ntoa(sin->sin_addr), asctime(timeinfo));

		this->sqlite.exec("INSERT INTO SecurityBlockedIp (ipAddress, blockedFrom, blockedTo) VALUES (%u, %d, %d)",
				htonl(sin->sin_addr.s_addr),
				(int) tNow,
				(int) tTo
				);
		firewallBlockAddress(sin);

		emailNotification(sin, failureCount, tNow, tTo);
	}
}

/**
 * Updates database when auth was successful
 *  - delete entries from SecurityLoginFailures
 *
 * This function is called from chan_sip
 *
 * @param sin
 * @param user
 */
void Security::authSuccess(const struct sockaddr_in *sin, const char *user)
{
	if (!this->loaded) {
		ast_log(AST_LOG_WARNING, "Security not loaded.\n");
		return;
	}
	LockUnique sqliteGuard(&this->sqliteLock.l);

	/* update SecurityLoginFailures */
	updateLoginFailures();

	/* because auth was successful, delete entries for this ip */
	this->sqlite.exec("DELETE FROM SecurityLoginFailures WHERE ipAddress = %u", htonl(sin->sin_addr.s_addr));
	if (this->sqlite.getChanges()) {
		ast_log(AST_LOG_WARNING, "[security] SIP: Authentication succeeded from host %s (uri:%s). Resetting attempts.\n", ast_inet_ntoa(sin->sin_addr), user);
	}

	firewallAllowAddress(sin);

	ast_debug(2, "SIP: Host %s authenticated (uri:%s).\n", ast_inet_ntoa(sin->sin_addr), user);
}

/**
 * Add blocked IP to the firewall ipset
 *  - timeout is set to actionBlockDuration/2
 */
void Security::firewallBlockAddress(const struct sockaddr_in * sin)
{
	std::string ip(ast_inet_ntoa(sin->sin_addr));

	ast_log(AST_LOG_NOTICE, "Blocking address \"%s\" on firewall...\n", ip.c_str());

	char tmp[1024];
	snprintf(tmp, sizeof (tmp), "/etc/boxinit.d/firewall block %s %d",
			ip.c_str(),
			this->actionBlockDuration / 2
			);

	ast_safe_system(tmp);

	ast_log(AST_LOG_NOTICE, "Blocking address \"%s\" on firewall... done(%d).\n", ip.c_str(), this->actionBlockDuration / 2);
}

/**
 * Add allowed IP to the firewall ipset
 */
void Security::firewallAllowAddress(const struct sockaddr_in * sin)
{
	std::string ip(ast_inet_ntoa(sin->sin_addr));

	if (this->firewallAllowedAddresses.find(ip) == this->firewallAllowedAddresses.end()) {
		this->firewallAllowedAddresses.insert(ip);

		ast_log(AST_LOG_NOTICE, "Allowing address \"%s\" on firewall...\n", ip.c_str());

		char tmp[1024];
		snprintf(tmp, sizeof (tmp), "/etc/boxinit.d/firewall allow %s",
				ip.c_str()
				);

		ast_safe_system(tmp);

		ast_log(AST_LOG_NOTICE, "Allowing address \"%s\" on firewall... done.\n", ip.c_str());
	}
}
